# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

from typing import Any, Dict, Optional

from comet_mpm.constants import EVENT_FEATURES, EVENT_PREDICTION
from comet_mpm.events.base_event import BaseEvent


class PredictionEvent(BaseEvent):
    """
    This class represents a single prediction event. Events are identified by the
    mandatory prediction_id parameter. Many events can have the same
    prediction_id. The MPM backend merges events with the same prediction_id
    automatically.
    Args:
        workspace: The project workspace.
        model_name: The name of model
        model_version: The version of model
        prediction_id: The unique prediction ID, could be provided by the
            framework, you or a random unique value could be provided like
            str(uuid4())
        input_features: If provided must be a flat Dictionary where the
            keys are the feature name and the value are native Python
            scalars, int, floats, booleans or strings. For example:
            `{“age”: 42, “income”: 42894.89}`
        output_features: A dictionary of output features
    """

    def __init__(
        self,
        workspace: str,
        model_name: str,
        model_version: str,
        prediction_id: str,
        input_features: Optional[Dict[str, Any]] = None,
        output_features: Optional[Dict[str, Any]] = None,
        timestamp: Optional[float] = None,
    ):
        super(PredictionEvent, self).__init__(
            workspace=workspace,
            model_name=model_name,
            model_version=model_version,
            prediction_id=prediction_id,
            timestamp=timestamp,
        )
        self.input_features = input_features
        self.output_features = output_features

    def _get_event_dict(self) -> Dict[str, Any]:
        event: Dict[str, Any] = {}
        if self.input_features is not None:
            event[EVENT_FEATURES] = self.input_features

        if self.output_features:
            event[EVENT_PREDICTION] = self.output_features

        return event



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import asyncio
import logging
from typing import Any, Optional, Tuple

import aiohttp

from ..batch_utils import Batch
from ..connection import send_asyncio_batch_requests
from ..events.label_event import LabelEvent
from ..events.prediction_event import PredictionEvent
from ..exceptions import AsyncioSenderNoConnected
from ..logging_messages import (
    ASYNCIO_SENDER_HARD_SHUTDOWN,
    BATCH_SENDING_EXCEPTION,
    LABEL_EVENT_SEND_ERROR,
    UNEXPECTED_EVENT_TYPE,
)
from .base import CLOSE_MESSAGE, BaseSender, batch_endpoint_url, labels_endpoint_url

LOGGER = logging.getLogger(__name__)


def get_asyncio_sender(
    api_key: str,
    server_address: str,
    max_batch_size: int,
    max_batch_time: int,
    batch_sending_timeout: int,
) -> "AsyncioSender":
    sender = AsyncioSender(
        api_key=api_key,
        server_address=server_address,
        max_batch_size=max_batch_size,
        max_batch_time=max_batch_time,
        batch_sending_timeout=batch_sending_timeout,
    )

    return sender


class AsyncioBackgroundSender:
    def __init__(
        self,
        api_key: str,
        server_address: str,
        queue: "asyncio.Queue[Any]",
        max_batch_size: int,
        max_batch_time: int,
        batch_sending_timeout: int,
    ):
        """A background coroutine that reads from the queue, add the items to the batch and sends
        them whenever the batch is ready to be sent. Can only be created only when an event loop is running.
        """
        self.api_key = api_key
        self.server_address = server_address
        self.session: aiohttp.ClientSession = aiohttp.ClientSession()
        self.queue: "asyncio.Queue[Any]" = queue
        self.max_batch_time = max_batch_time
        self.batch = Batch(max_batch_size, self.max_batch_time)
        self.batch_sending_timeout = batch_sending_timeout

        self.stop_processing = False

        # Pre-compute URLs
        self.batch_endpoint = batch_endpoint_url(self.server_address)
        self.labels_endpoint = labels_endpoint_url(self.server_address)

    async def run(self) -> None:
        while True:
            stop = await self._loop()

            if stop is True:
                break

        # We force stop the processing of the queue
        await self._check_batch_sending()
        return

    async def _loop(self) -> bool:
        try:
            data = await asyncio.wait_for(self.queue.get(), timeout=self.max_batch_time)
            if data is None:
                return False

            if data is CLOSE_MESSAGE:
                self.stop_processing = True
                return True

            if isinstance(data, PredictionEvent):
                self._process_prediction_event(data)
            elif isinstance(data, LabelEvent):
                await self._process_label_event(data)
            else:
                LOGGER.error(UNEXPECTED_EVENT_TYPE, data)
        except asyncio.TimeoutError:
            pass

        await self._check_batch_sending()

        return False

    async def _process_label_event(self, event: LabelEvent) -> None:
        try:
            await send_asyncio_batch_requests(
                session=self.session,
                url_endpoint=self.labels_endpoint,
                api_key=self.api_key,
                batch_sending_timeout=self.batch_sending_timeout,
                batch=event.create_event_dict(),
            )
        except Exception:
            LOGGER.error(LABEL_EVENT_SEND_ERROR, exc_info=True)

    def _process_prediction_event(self, event: PredictionEvent) -> None:
        to_send = event.create_event_dict()
        self.batch.append(to_send)

    async def _send_batch(self, batch: Any) -> None:
        async with aiohttp.ClientSession() as session:
            await send_asyncio_batch_requests(
                session=session,
                url_endpoint=self.batch_endpoint,
                api_key=self.api_key,
                batch_sending_timeout=self.batch_sending_timeout,
                batch=batch,
            )

    async def _check_batch_sending(self) -> None:
        # This should only be called by the background sender coroutine, if not the batch could
        # become corrupted as its access not protected by a lock
        try:
            if self.batch.should_be_uploaded(self.stop_processing):
                batch = self.batch.get_and_clear()

                await send_asyncio_batch_requests(
                    session=self.session,
                    url_endpoint=self.batch_endpoint,
                    api_key=self.api_key,
                    batch_sending_timeout=self.batch_sending_timeout,
                    batch=batch,
                )
        except Exception:
            LOGGER.error(BATCH_SENDING_EXCEPTION, exc_info=True)

    def close(self) -> None:
        """Force the AsyncioBackgroundSender to stop processing messages"""
        #
        self.stop_processing = True

        while True:
            try:
                data = self.queue.get_nowait()
                if data and data is not CLOSE_MESSAGE:
                    self._process_prediction_event(data)
            except asyncio.QueueEmpty:
                break

        asyncio.run(self._send_batch(self.batch.get_and_clear()))


class AsyncioSender(BaseSender):
    def __init__(
        self,
        api_key: str,
        server_address: str,
        max_batch_size: int,
        max_batch_time: int,
        batch_sending_timeout: int,
    ) -> None:
        self.sender_queue: Optional["asyncio.Queue[Any]"] = None
        # We need to wait for the asyncio event loop to be created
        self.background_sender: Optional[AsyncioBackgroundSender] = None
        self.background_task: Optional[asyncio.Task[None]] = None
        self.lock: Optional[asyncio.Lock] = None
        self.drain = False

        self.api_key = api_key
        self.server_address = server_address
        self.max_batch_time = max_batch_time
        self.max_batch_size = max_batch_size
        self.batch_sending_timeout = batch_sending_timeout

    def connect(self) -> None:
        if self.sender_queue is None:
            sender_queue: "asyncio.Queue[Any]" = asyncio.Queue()
            self.sender_queue = sender_queue
            self.background_sender = AsyncioBackgroundSender(
                api_key=self.api_key,
                server_address=self.server_address,
                queue=self.sender_queue,
                max_batch_size=self.max_batch_size,
                max_batch_time=self.max_batch_time,
                batch_sending_timeout=self.batch_sending_timeout,
            )
            self.background_sender.queue = self.sender_queue

        if self.background_task is None:
            assert self.background_sender is not None
            self.background_task = asyncio.create_task(self.background_sender.run())

        if self.lock is None:
            self.lock = asyncio.Lock()

    def _check_connected(self) -> Tuple[asyncio.Lock, "asyncio.Queue[Any]"]:
        if self.lock is None or self.sender_queue is None:
            raise AsyncioSenderNoConnected()

        return self.lock, self.sender_queue

    async def put(self, item: Any) -> None:
        lock, queue = self._check_connected()
        async with lock:
            if not self.drain:
                await queue.put(item)

    def close(self) -> None:
        """Called during atexit but users should call join manually. In last resort, synchronously
        dequeu items and send it by recreating an asyncio Event Loop
        """
        if self.drain is True:
            return None
        else:
            self.drain = True

            need_to_clean = False

            if self.sender_queue is not None and self.sender_queue.qsize() > 0:
                need_to_clean = True

            if (
                self.background_sender is not None
                and len(self.background_sender.batch) > 0
            ):
                need_to_clean = True

            if need_to_clean is False:
                return None
            else:
                LOGGER.warning(
                    ASYNCIO_SENDER_HARD_SHUTDOWN
                )  # TODO: Add link to documentation page
                self.drain = True

                # If the sender_queue is not None, the background sender shouldn't be None either
                assert self.background_sender is not None
                self.background_sender.close()

    async def join(self, timeout: int) -> None:
        """Inform the background sender to stops processing after receiving the sentinel object and
        wait for the background sender to stop
        """
        try:
            lock, queue = self._check_connected()
        except AsyncioSenderNoConnected:
            # The AsyncioSender was not properly initialized, ignore
            return

        async with lock:
            if self.drain is False:
                self.drain = True

                queue.put_nowait(CLOSE_MESSAGE)

                # TODO: Handle asyncio cleaning
                if self.background_task is not None:
                    await asyncio.wait_for(self.background_task, timeout)



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import logging
import queue
from threading import Lock, Thread
from typing import Any, Optional, Tuple

from ..batch_utils import ThreadSafeBatch
from ..connection import get_http_session, get_retry_strategy, send_batch_requests
from ..events.label_event import LabelEvent
from ..events.prediction_event import PredictionEvent
from ..logging_messages import (
    BATCH_SENDING_EXCEPTION,
    LABEL_EVENT_SEND_ERROR,
    UNEXPECTED_EVENT_TYPE,
)
from .base import CLOSE_MESSAGE, BaseSender, batch_endpoint_url, labels_endpoint_url

LOGGER = logging.getLogger(__name__)


class ThreadBackgroundSender:
    def __init__(
        self,
        api_key: str,
        server_address: str,
        sender_queue: "queue.Queue[Any]",
        max_batch_size: int,
        max_batch_time: int,
        batch_sending_timeout: int,
    ):
        self.api_key = api_key
        self.server_address = server_address
        self.sender_queue = sender_queue
        self.max_batch_time = max_batch_time
        self.batch = ThreadSafeBatch(max_batch_size, self.max_batch_time)
        self.session = get_http_session(retry_strategy=get_retry_strategy())
        self.batch_sending_timeout = batch_sending_timeout

        self.stop_processing = False

        # Pre-compute URLs
        self.batch_endpoint = batch_endpoint_url(self.server_address)
        self.labels_endpoint = labels_endpoint_url(self.server_address)

    def run(self) -> None:
        """This is meant to be run in an independent Thread"""
        while self.stop_processing is False:
            stop = self._loop()

            if stop is True:
                break

        # We force stop the processing of the queue
        self._check_batch_sending()
        return

    def _loop(self) -> bool:
        try:
            data = self.sender_queue.get(block=True, timeout=self.max_batch_time)
            if data is None:
                return False

            if data is CLOSE_MESSAGE:
                self.stop_processing = True
                return True

            if isinstance(data, PredictionEvent):
                self._process_prediction_event(event=data)
            elif isinstance(data, LabelEvent):
                self._process_label_event(event=data)
            else:
                LOGGER.error(UNEXPECTED_EVENT_TYPE, data)
        except queue.Empty:
            pass

        self._check_batch_sending()

        return False

    def _process_label_event(self, event: LabelEvent) -> None:
        try:
            send_batch_requests(
                self.session,
                self.labels_endpoint,
                api_key=self.api_key,
                batch=event.create_event_dict(),
                batch_sending_timeout=self.batch_sending_timeout,
            )
        except Exception:
            LOGGER.error(LABEL_EVENT_SEND_ERROR, exc_info=True)

    def _process_prediction_event(self, event: PredictionEvent) -> None:
        to_send = event.create_event_dict()
        self.batch.append(to_send)

    def _check_batch_sending(self) -> None:
        try:
            if self.batch.should_be_uploaded(self.stop_processing):
                batch = self.batch.get_and_clear()

                send_batch_requests(
                    self.session,
                    self.batch_endpoint,
                    api_key=self.api_key,
                    batch=batch,
                    batch_sending_timeout=self.batch_sending_timeout,
                )
        except Exception:
            LOGGER.error(BATCH_SENDING_EXCEPTION, exc_info=True)

    def close(self) -> None:
        """For the BackgroundSender to stop processing messages"""
        self.stop_processing = True
        # This shouldn't happen outside of the background sender thread
        self._check_batch_sending()

        self.session.close()


class ThreadSender(BaseSender):
    def __init__(
        self,
        api_key: str,
        server_address: str,
        max_batch_size: int,
        max_batch_time: int,
        batch_sending_timeout: int,
    ) -> None:
        self.lock = Lock()
        self.sender_queue: Optional["queue.Queue[Any]"] = None
        self.background_sender: Optional[ThreadBackgroundSender] = None
        self.background_thread: Optional[Thread] = None
        self.drain = False

        self.api_key = api_key
        self.server_address = server_address
        self.max_batch_time = max_batch_time
        self.max_batch_size = max_batch_size
        self.batch_sending_timeout = batch_sending_timeout

    def _prepare(self) -> "queue.Queue[Any]":
        """We need to create the background thread on the first request to support application
        pre-loading.
        """
        if self.sender_queue is None:
            sender_queue: "queue.Queue[Any]" = queue.Queue()
            self.sender_queue = sender_queue
            self.background_sender = ThreadBackgroundSender(
                api_key=self.api_key,
                server_address=self.server_address,
                sender_queue=self.sender_queue,
                max_batch_size=self.max_batch_size,
                max_batch_time=self.max_batch_time,
                batch_sending_timeout=self.batch_sending_timeout,
            )

        if self.background_thread is None:
            assert self.background_sender is not None
            self.background_thread = Thread(
                target=self.background_sender.run, daemon=True, name="ThreadSender"
            )
            self.background_thread.start()

        assert self.sender_queue is not None
        return self.sender_queue

    def put(self, item: Any) -> None:
        with self.lock:
            if not self.drain:

                if self.sender_queue is None:
                    sender_queue = self._prepare()
                else:
                    sender_queue = self.sender_queue

                sender_queue.put(item)

    def close(self) -> None:
        with self.lock:
            self.drain = True

        if self.sender_queue is not None:
            self.sender_queue.put(CLOSE_MESSAGE)

        if self.background_thread is not None:
            self.background_thread.join(10)

        if self.background_sender is not None:
            self.background_sender.close()

    def join(self, timeout: int) -> None:
        """There is no easy way to plug a shutdown callback with Flask, the "normal" cleaning
        process for Flask is to use the close method"""
        return None

    def connect(self) -> None:
        """There is no easy way nor need to plug a startup callback as we can create everything and
        do the handshake during Python import"""
        return None


def get_thread_sender(
    api_key: str,
    server_address: str,
    max_batch_size: int,
    max_batch_time: int,
    batch_sending_timeout: int,
) -> "ThreadSender":
    sender = ThreadSender(
        api_key=api_key,
        server_address=server_address,
        max_batch_size=max_batch_size,
        max_batch_time=max_batch_time,
        batch_sending_timeout=batch_sending_timeout,
    )

    return sender


def _get_thread_background_sender(
    api_key: str,
    server_address: str,
    max_batch_size: int,
    max_batch_time: int,
    batch_sending_timeout: int,
) -> Tuple["queue.Queue[Any]", "ThreadBackgroundSender"]:
    sender_queue: queue.Queue[Any] = queue.Queue()
    background_sender = ThreadBackgroundSender(
        api_key=api_key,
        server_address=server_address,
        sender_queue=sender_queue,
        max_batch_size=max_batch_size,
        max_batch_time=max_batch_time,
        batch_sending_timeout=batch_sending_timeout,
    )

    return sender_queue, background_sender



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2015-2021 Comet ML INC
#  This file can not be copied and/or distributed
#  without the express permission of Comet ML Inc.
# *******************************************************

from threading import Lock
from time import monotonic
from typing import Any, List


class Batch(object):
    """
    The Batch object contains a list of anything and manage the size of the batch, isolating the
    logic about the max size and max time for a batch
    """

    def __init__(self, max_batch_size: int, max_batch_time: int) -> None:
        self.batch: List[Any] = []
        self.last_time_created = monotonic()
        self.critical = False

        self.max_batch_size = max_batch_size
        self.max_batch_time = max_batch_time

    def append(self, batch_item: Any) -> None:
        if len(self.batch) == 0:
            self.last_time_created = monotonic()

        self.batch.append(batch_item)

    def __len__(self) -> int:
        return len(self.batch)

    def should_be_uploaded(self, flush: bool) -> bool:
        """Check if the current batch should be uploaded according to both max_batch_size and
        max_batch_item. If the batch is empty. If flush is True, always return True except if the batch is
        empty.
        """
        if len(self.batch) == 0:
            return False

        if flush:
            return flush

        duration_since_last_created = monotonic() - self.last_time_created

        too_many_items = len(self.batch) >= self.max_batch_size
        too_old_batch = duration_since_last_created >= self.max_batch_time

        return too_many_items or too_old_batch

    def clear(self) -> None:
        self.batch = []

    def get_and_clear(self) -> List[Any]:
        batch = self.batch
        self.batch = []
        return batch


class ThreadSafeBatch(Batch):
    """
    The Batch object contains a list of anything and manage the size of the batch, isolating the
    logic about the max size and max time for a batch
    """

    def __init__(self, max_batch_size: int, max_batch_time: int) -> None:
        super().__init__(max_batch_size=max_batch_size, max_batch_time=max_batch_time)
        self.lock = Lock()

    def append(self, batch_item: Any) -> None:
        with self.lock:
            return super().append(batch_item=batch_item)

    def __len__(self) -> int:
        with self.lock:
            return super().__len__()

    def should_be_uploaded(self, flush: bool) -> bool:
        with self.lock:
            return super().should_be_uploaded(flush=flush)

    def clear(self) -> None:
        with self.lock:
            return super().clear()

    def get_and_clear(self) -> List[Any]:
        with self.lock:
            return super().get_and_clear()



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021-2023 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************
import abc
import calendar
from datetime import datetime
from typing import Any, Dict, Optional

from comet_mpm.constants import (
    EVENT_MODEL_NAME,
    EVENT_MODEL_VERSION,
    EVENT_PREDICTION_ID,
    EVENT_TIMESTAMP,
    EVENT_WORKSPACE_NAME,
)


class BaseEvent(abc.ABC):
    def __init__(
        self,
        workspace: str,
        model_name: str,
        model_version: str,
        prediction_id: str,
        timestamp: Optional[float] = None,
    ):
        self.workspace = workspace
        self.model_name = model_name
        self.model_version = model_version
        self.prediction_id = prediction_id

        self.timestamp = local_timestamp(timestamp)

    def create_event_dict(self) -> Dict[str, Any]:
        base_dict = {
            EVENT_MODEL_NAME: self.model_name,
            EVENT_MODEL_VERSION: self.model_version,
            EVENT_WORKSPACE_NAME: self.workspace,
            EVENT_PREDICTION_ID: self.prediction_id,
            EVENT_TIMESTAMP: self.timestamp,
        }
        base_dict.update(self._get_event_dict())
        return base_dict

    @abc.abstractmethod
    def _get_event_dict(self) -> Dict[str, Any]:
        ...


def local_timestamp(timestamp: Optional[float] = None) -> int:
    """Return a timestamp in a format expected by the backend (milliseconds)"""
    if timestamp is None:
        now = datetime.utcnow()
        timestamp_in_seconds = calendar.timegm(now.timetuple()) + (
            now.microsecond / 1e6
        )
    else:
        timestamp_in_seconds = timestamp
    timestamp_in_milliseconds = int(timestamp_in_seconds * 1000)
    return timestamp_in_milliseconds



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************
import json
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import numpy.typing as npt
import pandas as pd
import pandas.api.types as types
from typing_extensions import TypedDict

from ._json.numpy_encoder import NumpyEncoder
from .logging_messages import ERROR_QUANTILES_INVALID_VALUE, ERROR_QUANTILES_LIST_EMPTY

"""This module holds utilities to process data series and data frames"""


class CategoricalSeriesSummary(TypedDict, total=False):
    """The summary of the categorical series"""

    missing_values: int
    count_distribution: List[Tuple[Union[int, float, str, bool], int]]


class NumericalDistributionStats(TypedDict):
    """Represents statistics about numerical distribution in series"""

    quantiles: List[float]
    values: List[float]


class NumericalSeriesSummary(TypedDict, total=False):
    """The summary of the numerical series"""

    missing_values: int
    numerical_distribution: NumericalDistributionStats


class DataFrameDistribution(TypedDict):
    """The summary of the data series distributions in the DataFrame"""

    general_distribution: Dict[
        str, Union[CategoricalSeriesSummary, NumericalSeriesSummary]
    ]


def to_json(
    distribution: Union[
        DataFrameDistribution, NumericalSeriesSummary, CategoricalSeriesSummary
    ]
) -> str:
    """Dumps provided distribution to JSON
    Args:
        distribution: Union[DataFrameDistribution, NumericalSeriesSummary, CategoricalSeriesSummary]
            The distribution to be dumped to JSON
    """
    return json.dumps(distribution, cls=NumpyEncoder)


def categorical_series_summary(
    series: pd.Series, max_distinct_values: int = 100, sort: bool = True
) -> CategoricalSeriesSummary:
    """Computes summary of categorical values from specified series.
    Args:
        series : pandas.core.series.Series
            The series to be evaluated
        max_distinct_values : int, default 100
            The maximal allowed number of distinct values per series.
        sort : bool, default True
            Sort by frequencies.
    Raises:
        TypeError : if provided series has wrong data type. Only int, float, str, bool categorical values are supported.
        ValueError : if there are more than max_distinct_values distinct values per series
    """
    # count all missing values
    missing_values = series.isna().sum()

    # check that maximal distinct values within range (including None)
    unique = series.nunique(dropna=False)
    if unique > max_distinct_values:
        raise ValueError(
            "The series has too many distinct values: %d. Max %d of distinct values are allowed."
            % (unique, max_distinct_values)
        )

    # drop missing values and check that dtype of remaining values are supported
    series = series.dropna(axis="rows")
    if not (
        series.apply(lambda x: types.is_number(x) or types.is_bool(x)).all()
        or types.is_string_dtype(series)
    ):
        raise TypeError("Unsupported series data type: %s" % series.dtype)

    # count all categories
    counts = series.value_counts(dropna=True, sort=sort)
    count_distribution = [(k, v) for k, v in counts.items()]

    return CategoricalSeriesSummary(
        missing_values=missing_values, count_distribution=count_distribution
    )


def numerical_series_summary(
    series: pd.Series, quantiles: List[float]
) -> NumericalSeriesSummary:
    """Computes summary of the provided numerical series
    Args:
        series: pandas.core.series.Series
            The series to be evaluated
        quantiles: List[float]
            The list of quantiles to compute
    Raises:
        TypeError : if provided series has wrong data type. Only numerical data types are supported.
    """
    if not types.is_numeric_dtype(series) or types.is_bool_dtype(series):
        raise TypeError("Unsupported series data type: %s" % series.dtype)

    # check that quantile values are valid
    _validate_quantiles(quantiles)

    # count all missing values
    na_positions = series.isna()
    missing_values = na_positions.sum()

    # calculate statistics if appropriate
    if not na_positions.all():
        values_raw = series.dropna().quantile(quantiles)
        values = [v for i, v in values_raw.items()]
        stats = NumericalDistributionStats(
            quantiles=quantiles,
            values=values,
        )

        return NumericalSeriesSummary(
            missing_values=missing_values, numerical_distribution=stats
        )
    else:
        return NumericalSeriesSummary(missing_values=missing_values)


def _compute_distribution(
    df: pd.DataFrame,
    numerical_features: Optional[List[str]] = None,
    numerical_quantiles: Optional[List[float]] = None,
    categorical_features: Optional[List[str]] = None,
) -> DataFrameDistribution:
    """Computes data series distribution along the columns in the provided data frame.

    Args:
        df: pandas.DataFrame
            The DataFrame to be evaluated.
        numerical_features: Optional[List[str]]
            The list with column names to calculate numerical distribution summary
        numerical_quantiles: Optional[List[float]]
            The list of quantiles to be calculated from numerical features. It must be present if numerical_features
            columns are present.
        categorical_features: Optional[List[str]]
            The list with column names to calculate categorical distribution summary
    Raises:
        TypeError: if when summary requested for specific column with wrong data type, e.g. numerical for column
            holding string data type series.
        ValueError: if requested summary for the column which doesn't exists in the data frame or if numerical_quantiles
            has wrong values. For details see: numerical_series_summary()
    """
    # check that all requested features are present in the data frame
    cols: List[str] = []
    if numerical_features is not None:
        cols = cols + numerical_features
        _validate_quantiles(numerical_quantiles)
    if categorical_features is not None:
        cols = cols + categorical_features

    for col in cols:
        if col not in df.columns:
            raise ValueError(
                "The requested feature column: %s is not found in the DataFrame" % col
            )

    # check that all requested features exactly match the number of feature columns in the data frame
    if len(cols) != len(df.columns):
        raise ValueError(
            "The DataFrame has more feature columns than number of features requested"
        )

    distributions: Dict[
        str, Union[CategoricalSeriesSummary, NumericalSeriesSummary]
    ] = {}
    # calculate numerical distributions
    if numerical_features is not None:
        assert numerical_quantiles is not None  # it must not be None at this stage
        for feature in numerical_features:
            distributions[feature] = numerical_series_summary(
                df[feature], quantiles=numerical_quantiles
            )

    # calculate categorical distributions
    if categorical_features is not None:
        for feature in categorical_features:
            distributions[feature] = categorical_series_summary(df[feature])

    return DataFrameDistribution(general_distribution=distributions)


def _validate_quantiles(quantiles: Optional[List[float]]) -> None:
    """Validates if provides list of quantile values is correct"""
    if quantiles is None or len(quantiles) == 0:
        raise ValueError(ERROR_QUANTILES_LIST_EMPTY)
    arr = np.array(quantiles)  # type: npt.NDArray[np.float64]
    min_quantile, max_quantile = np.amin(arr), np.amax(arr)
    if min_quantile < 0 or max_quantile > 1:
        raise ValueError(ERROR_QUANTILES_INVALID_VALUE % (min_quantile, max_quantile))



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import logging
from typing import Any, Optional
from urllib.parse import urljoin, urlparse, urlunparse

import aiohttp
from requests import Session
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from .logging_messages import BATCH_SENDING_ERROR

LOGGER = logging.getLogger(__name__)

MPM_BASE_PATH = "/mpm/"
REST_API_BASE_PATH = "/api/v2/"


def get_retry_strategy() -> Retry:
    # The total backoff sleeping time is computed like that:
    # backoff = 2
    # total = 3
    # s = lambda b, i: b * (2 ** (i - 1))
    # sleep = sum(s(backoff, i) for i in range(1, total + 1))

    return Retry(
        total=3,
        backoff_factor=2,
        allowed_methods=None,
    )  # Will wait up to 24s


def get_http_session(retry_strategy: Optional[Retry] = None) -> Session:
    session = Session()

    # Setup retry strategy if asked
    if retry_strategy is not None:
        session.mount("http://", HTTPAdapter(max_retries=retry_strategy))
        session.mount("https://", HTTPAdapter(max_retries=retry_strategy))

    return session


def send_batch_requests(
    session: Session,
    url_endpoint: str,
    api_key: str,
    batch_sending_timeout: int,
    batch: Any,
) -> None:
    try:
        headers = {"Authorization": api_key}
        response = session.post(
            url_endpoint, headers=headers, json=batch, timeout=batch_sending_timeout
        )
        response.raise_for_status()
        # TODO: Process response
    except Exception:
        LOGGER.error(BATCH_SENDING_ERROR, exc_info=True)


async def send_asyncio_batch_requests(
    session: aiohttp.ClientSession,
    url_endpoint: str,
    api_key: str,
    batch_sending_timeout: int,
    batch: Any,
) -> None:
    try:
        headers = {"Authorization": api_key}
        async with session.post(
            url_endpoint, headers=headers, json=batch, timeout=batch_sending_timeout
        ) as response:
            response.raise_for_status()
            # TODO: Process response
            _ = await response.text()
    except Exception:
        LOGGER.error(BATCH_SENDING_ERROR, exc_info=True)


def sanitize_url(url: str) -> str:
    """Sanitize an URL, checking that it is a valid URL and ensure it contains an ending slash /"""
    parts = urlparse(url)
    scheme, netloc, path, params, query, fragment = parts

    # TODO: Raise an exception if params, query and fragment are not empty?

    # Ensure the leading slash
    if path and not path.endswith("/"):
        path = path + "/"
    elif not path and not netloc.endswith("/"):
        netloc = netloc + "/"

    return urlunparse((scheme, netloc, path, params, query, fragment))


def url_join(base: str, *parts: str) -> str:
    """Given a base and url parts (for example [workspace, project, id]) returns a full URL"""
    # TODO: Enforce base to have a scheme and netloc?
    result = base

    for part in parts[:-1]:
        if not part.endswith("/"):
            raise ValueError("Intermediary part not ending with /")

        result = urljoin(result, part)

    result = urljoin(result, parts[-1])

    return result



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021-2023 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************
from typing import Any, Dict

from comet_mpm.constants import EVENT_PREDICTION_VALUE
from comet_mpm.events.base_event import BaseEvent


class LabelEvent(BaseEvent):
    """
    This class represents a single ground truth label.
    Args:
        workspace: The project workspace.
        model_name: The name of model
        model_version: The version of model
        prediction_id: The unique prediction ID, could be provided by the
            framework, you or a random unique value could be provided like
            str(uuid4())
        label: the ground truth value for a prediction
    """

    def __init__(
        self,
        workspace: str,
        model_name: str,
        model_version: str,
        prediction_id: str,
        label: Any,
    ):
        super(LabelEvent, self).__init__(
            workspace=workspace,
            model_name=model_name,
            model_version=model_version,
            prediction_id=prediction_id,
        )
        self.label = label

    def _get_event_dict(self) -> Dict[str, Any]:
        return {EVENT_PREDICTION_VALUE: self.label}



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import asyncio as asyncio_module
import atexit
import logging
import os
from typing import Any, Awaitable, Dict, Iterable, List, Optional, Union

from . import constants, logging_messages, optional_update
from .connection import MPM_BASE_PATH, REST_API_BASE_PATH, sanitize_url, url_join
from .events import events_from_dataframe
from .events.label_event import LabelEvent
from .events.prediction_event import PredictionEvent
from .sender import get_sender
from .settings import MPMSettings, get_model

LOGGER = logging.getLogger(__name__)

LogEventsResult = Union[List[Any], Awaitable[List[Any]]]


class CometMPM:
    """
    The Comet MPM class is used to upload a model's input and output features to MPM
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        workspace_name: Optional[str] = None,
        model_name: Optional[str] = None,
        model_version: Optional[str] = None,
        disabled: Optional[bool] = None,
        asyncio: bool = False,
        max_batch_size: Optional[int] = None,
        max_batch_time: Optional[int] = None,
    ):
        """
        Creates the Comet MPM Event logger object.
        Args:
            api_key: The Comet API Key
            workspace_name: The Comet Workspace Name of the model
            model_name: The Comet Model Name of the model
            model_version: The Comet Model Version of the model
            disabled: If set to True, CometMPM will not send anything to the backend.
            asyncio: Set to True if you are using an Asyncio-based framework like FastAPI.
            max_batch_size: Maximum number of MPM events sent in a batch, can also be configured using the environment variable MPM_MAX_BATCH_SIZE.
            max_batch_time: Maximum time before a batch of events is submitted to MPM, can also be configured using the environment variable MPM_MAX_BATCH_SIZE.
        """

        settings_user_values = {}  # type: Dict[str, str|int]
        optional_update.update(
            settings_user_values,
            {
                "api_key": api_key,
                "mpm_model_name": model_name,
                "mpm_model_version": model_version,
                "mpm_workspace_name": workspace_name,
                "mpm_max_batch_size": max_batch_size,
                "mpm_max_batch_time": max_batch_time,
            },
        )

        self._settings = get_model(
            MPMSettings,
            **settings_user_values,
        )
        if disabled:
            self.disabled = disabled  # type: bool
        else:
            self.disabled = bool(os.getenv("COMET_MPM_DISABLED"))
        self._asyncio = asyncio

        self._mpm_url = url_join(sanitize_url(self._settings.url), MPM_BASE_PATH)
        self._api_url = url_join(sanitize_url(self._settings.url), REST_API_BASE_PATH)

        if self.disabled:
            self._sender = None
        else:
            self._sender = get_sender(
                api_key=self._settings.api_key,
                server_address=self._mpm_url,
                max_batch_size=self._settings.mpm_max_batch_size,
                max_batch_time=self._settings.mpm_max_batch_time,
                asyncio=self._asyncio,
                batch_sending_timeout=self._settings.mpm_batch_sending_timeout,
            )

            atexit.register(self._on_end)

        # TODO: Do handshake

    def log_event(
        self,
        prediction_id: str,
        input_features: Optional[Dict[str, Any]] = None,
        output_value: Optional[Any] = None,
        output_probability: Optional[Any] = None,
        output_features: Optional[Dict[str, Any]] = None,
        timestamp: Optional[float] = None,
    ) -> Optional[Awaitable[None]]:
        """
        Asynchronously log a single event to MPM. Events are identified by the
        mandatory prediction_id parameter. You can send multiple events with
        the same prediction_id, and the events will be automatically merged
        on the backend side.

        Args:
            prediction_id: The unique prediction ID. It can be provided by the
                framework, you, or a random unique value such as str(uuid4()).
            input_features: If provided, it must be a flat dictionary where the
                keys are the feature names, and the values are native Python
                scalars, such as integers, floats, booleans, or strings. For
                example: `{"age": 42, "income": 42894.89}`.
            output_value: The prediction as a native Python scalar, such as an
                 integer, float, boolean, or string.
            output_probability: If provided, it must be a float between 0 and
                 1, indicating the model's confidence in the prediction.
            output_features: A dictionary of output features.
            timestamp: An optional timestamp to associate with the event
                (seconds since epoch in UTC timezone). If not provided, the
                 current time will be used."""
        if self.disabled:
            if self._asyncio is False:
                return None
            else:
                return asyncio_module.sleep(0)

        output_features = _handle_event_output_features(
            output_value, output_probability, output_features
        )

        event = PredictionEvent(
            workspace=self._settings.mpm_workspace_name,
            model_name=self._settings.mpm_model_name,
            model_version=self._settings.mpm_model_version,
            prediction_id=prediction_id,
            input_features=input_features,
            output_features=output_features,
            timestamp=timestamp,
        )
        return self._log_event(event)

    def log_label(self, prediction_id: str, label: Any) -> Optional[Awaitable[None]]:
        """
        Send an MPM event containing the ground truth value for a prediction whose input and output
        features are already stored in Comet.
        Args:
            prediction_id: The unique prediction ID
            label: The ground truth value for the prediction
        """
        if self.disabled:
            if self._asyncio is False:
                return None
            else:
                return asyncio_module.sleep(0)

        event = LabelEvent(
            workspace=self._settings.mpm_workspace_name,
            model_name=self._settings.mpm_model_name,
            model_version=self._settings.mpm_model_version,
            prediction_id=prediction_id,
            label=label,
        )
        return self._log_event(event)

    def log_dataframe(  # type: ignore[no-untyped-def]
        self,
        dataframe,
        prediction_id_column: str,
        feature_columns: Optional[List[str]] = None,
        output_value_column: Optional[str] = None,
        output_probability_column: Optional[str] = None,
        output_features_columns: Optional[List[str]] = None,
        timestamp_column: Optional[str] = None,
    ) -> LogEventsResult:
        """
        This function logs each row of a Pandas DataFrame as an MPM event. The
        events are structured as described in the [log_event](#cometmpmlog_event)
        method, so please refer to it for full context.

        Args:
            dataframe: The Pandas DataFrame to be logged.
            prediction_id_column: This column should contain the prediction_id values for the
                events.
            feature_columns: If provided, these columns will be used as the input_features
                for the events.
            output_features_columns: If provided, these columns will be used as the output_features for the events.
            output_value_column: If provided, this column will be used as the output_value for the
                events.
            output_probability_column: If provided, this column will be used as the
                output_probability for the events.
            timestamp_column: If provided, this column will be used as the timestamp (seconds since
                epoch start in UTC timezone) for the events.
        """
        events = events_from_dataframe.generate(
            workspace=self._settings.mpm_workspace_name,
            model_name=self._settings.mpm_model_name,
            model_version=self._settings.mpm_model_version,
            dataframe=dataframe,
            prediction_id_column=prediction_id_column,
            feature_columns=feature_columns,
            output_features_columns=output_features_columns,
            output_value_column=output_value_column,
            output_probability_column=output_probability_column,
            timestamp_column=timestamp_column,
        )

        return self._log_events(events)

    def connect(self) -> None:
        """
        When using CometMPM in asyncio mode, this coroutine needs to be awaited
        at the server start.
        """
        if self._sender is not None:
            self._sender.connect()

    def join(self, timeout: Optional[int] = None) -> Optional[Awaitable[None]]:
        """
        When using CometMPM in asyncio mode, this coroutine needs to be awaited
        at the server stop.
        """
        if timeout is None:
            timeout = self._settings.mpm_join_timeout

        if not self.disabled:
            assert self._sender is not None
            return self._sender.join(timeout)
        else:
            if self._asyncio is False:
                return None
            else:
                return asyncio_module.sleep(0)

    def _on_end(self) -> None:
        if not self.disabled:
            assert self._sender is not None
            self._sender.close()

    def _log_events(self, events: Iterable[PredictionEvent]) -> LogEventsResult:
        results = []
        for event in events:
            result = self._log_event(event)
            if result is not None:
                results.append(result)

        if self._asyncio and len(results) > 0:
            return asyncio_module.gather(*results)

        return results

    def _log_event(
        self, event: Union[PredictionEvent, LabelEvent]
    ) -> Optional[Awaitable[None]]:
        assert self._sender is not None
        return self._sender.put(event)


def _handle_event_output_features(
    output_value: Any,
    output_probability: Any,
    output_features: Optional[Dict[str, Any]],
) -> Optional[Dict[str, Any]]:
    event_output_features: Optional[Dict[str, Any]]

    if (
        output_value is not None or output_probability is not None
    ) and output_features is None:
        LOGGER.warning(
            logging_messages.DEPRECATED_OUTPUT_VALUE_AND_PROBABILITY_WITHOUT_FEATURES
        )

        event_output_features = {}
        if output_value is not None:
            event_output_features[constants.EVENT_PREDICTION_VALUE] = output_value

        if output_probability is not None:
            event_output_features[
                constants.EVENT_PREDICTION_PROBABILITY
            ] = output_probability
    elif (
        output_value is not None and output_probability is not None
    ) and output_features is not None:
        LOGGER.warning(
            logging_messages.DEPRECATED_OUTPUT_VALUE_AND_PROBABILITY_WITH_FEATURES
        )

        event_output_features = {
            constants.EVENT_PREDICTION_VALUE: output_value,
            constants.EVENT_PREDICTION_PROBABILITY: output_probability,
        }
        for key in output_features:
            event_output_features[key] = output_features[key]
    else:
        event_output_features = output_features

    return event_output_features



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

from abc import ABCMeta
from typing import Any, Awaitable, Optional
from urllib.parse import urljoin


def batch_endpoint_url(server_address: str) -> str:
    # TODO: Check that server_address finishes with a "/"
    return urljoin(server_address, "events/batch")


def labels_endpoint_url(server_address: str) -> str:
    return urljoin(server_address, "labels")


CLOSE_MESSAGE = object()


class BaseSender(metaclass=ABCMeta):
    def connect(self) -> None:
        ...  # pragma: no cover

    def put(self, item: Any) -> Optional[Awaitable[None]]:
        ...  # pragma: no cover

    def close(self) -> None:
        ...  # pragma: no cover

    def join(self, timeout: int) -> Optional[Awaitable[None]]:
        ...  # pragma: no cover



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************


import argparse
import sys


def main() -> int:
    """Console script for comet_python_mpm_sdk."""
    parser = argparse.ArgumentParser()
    parser.add_argument("_", nargs="*")
    args = parser.parse_args()

    print("Arguments: " + str(args._))
    print(
        "Replace this message by putting your code into "
        "comet_python_mpm_sdk.cli.main"
    )
    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************
EVENT_PREDICTION_ID = "predictionId"
EVENT_TIMESTAMP = "timestamp"
EVENT_FEATURES = "features"
EVENT_PREDICTION = "prediction"
EVENT_PREDICTION_VALUE = "value"
EVENT_PREDICTION_PROBABILITY = "probability"
EVENT_WORKSPACE_NAME = "workspaceName"
EVENT_MODEL_NAME = "modelName"
EVENT_MODEL_VERSION = "modelVersion"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2015-2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

from pathlib import Path

from setuptools import find_packages, setup

requirements = [
    "pandas",
    "typing_extensions>=3.7.4",
    "aiohttp",
    "pydantic<2.0.0",
    "requests",
]

# read the contents of your PACKAGE file

this_directory = Path(__file__).parent
long_description = (this_directory / "PACKAGE.md").read_text()


setup(
    author="Comet ML Inc.",
    author_email="mail@comet.ml",
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3",
    ],
    description="Comet MPM SDK",
    install_requires=requirements,
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="comet_mpm",
    name="comet_mpm",
    packages=find_packages("src"),
    package_dir={"": "src"},
    url="https://www.comet.ml",
    version="0.5.0",
    zip_safe=False,
    license="Proprietary",
)



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

BATCH_SENDING_ERROR = "Error sending batch data to the backend"

BATCH_SENDING_EXCEPTION = "Unknown exception happened when sending data to the backend"

LABEL_EVENT_SEND_ERROR = "Error sending label to the backend"

GEVENT_NOT_SUPPORTED = (
    "Using the MPM SDK with gevent is not officially supported, data loss might occurs"
)

ASYNCIO_SENDER_HARD_SHUTDOWN = "There are still some data in-flight and might be missing. Make sure to call CometMPM.join() manually to guarantee that all data has been sent to the Backend"

ERROR_QUANTILES_LIST_EMPTY = (
    "Error cannot compute distribution summary because the quantiles list is empty."
)
ERROR_QUANTILES_INVALID_VALUE = (
    "The quantile must be within [0,1]. The current min value %f and max %f"
)

LOGGING_DIRECTORY_CREATION_ERROR = "Cannot create path to log file %r"

LOGGING_LOG_FILE_OPEN_ERROR = "Cannot open log file %r; file logging is disabled"

INVALID_SETTING_DEFAULT_FALLBACK = "Some MPM settings values were invalid, default values were used instead, see below for the list"

UNEXPECTED_EVENT_TYPE = "Unexpected event type %r"

DEPRECATED_OUTPUT_VALUE_AND_PROBABILITY_WITHOUT_FEATURES = 'You are using the deprecated fields `output_value` and `output_probability`,you should specify `output_features={"value": "<>", "probability": "<>"}`'

DEPRECATED_OUTPUT_VALUE_AND_PROBABILITY_WITH_FEATURES = 'You are using the deprecated fields `output_value` and `output_probability`, you should specify `output_features={"value": "<>", "probability": "<>", ...}`'

DEPRECATED_DATAFRAME_OUTPUT_VALUE_AND_PROBABILITY = "You are using the deprecated fields `output_value_column` and `output_probability_column`,you should specify `output_features_columns with value and probability columns`"



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import getpass
import logging
import os
import re
from datetime import datetime
from typing import Optional

from .logging_messages import (
    LOGGING_DIRECTORY_CREATION_ERROR,
    LOGGING_LOG_FILE_OPEN_ERROR,
)
from .settings import LogLevel, LogSettings, get_model

LOGGER = logging.getLogger(__name__)

MSG_FORMAT = "COMET %(levelname)s: %(message)s"

FILE_MSG_FORMAT = "[%(process)d-%(processName)s:%(thread)d] %(relativeCreated)d COMET %(levelname)s [%(filename)s:%(lineno)d]: %(message)s"


def get_user() -> str:
    try:
        return getpass.getuser()
    except KeyError:
        return "unknown"


def _make_valid(s: str) -> str:
    s = str(s).strip().replace(" ", "_")
    return re.sub(r"(?u)[^-\w.]", "", s)


def expand_log_file_path(log_file_path: Optional[str]) -> Optional[str]:
    """
    Expand patterns in the file logging path.

    Allowed patterns:
        * {datetime}
        * {pid}
        * {project}
        * {user}
    """

    if log_file_path is None:
        return None

    user = _make_valid(get_user())

    patterns = {
        "datetime": datetime.now().strftime("%Y%m%d-%H%M%S"),
        "pid": os.getpid(),
        "user": user,
    }

    try:
        return log_file_path.format(**patterns)
    except KeyError:
        LOGGER.info(
            "Invalid logging file pattern: '%s'; ignoring",
            log_file_path,
            exc_info=True,
        )
        return log_file_path


class CometLoggingConfig(object):
    def __init__(self, settings: LogSettings) -> None:
        self.root = logging.getLogger("comet_mpm")
        logger_level = logging.CRITICAL

        # Don't send comet-mpm to the application logger
        self.root.propagate = settings.mpm_logging_propagate

        # Add handler for console, basic INFO:
        self.console_handler = logging.StreamHandler()

        logging_console_level = settings.mpm_logging_console
        self.console_formatter = logging.Formatter(MSG_FORMAT)

        self.console_handler.setLevel(logging_console_level)
        self.console_handler.setFormatter(self.console_formatter)
        self.root.addHandler(self.console_handler)

        logger_level = min(logger_level, self.console_handler.level)

        # The std* logger might conflicts with the logging if a log record is
        # emitted for each WS message as it would results in an infinite loop. To
        # avoid this issue, all log records after the creation of a message should
        # be at a level lower than info as the console handler is set to info
        # level.

        # Add an additional file handler
        log_file_path = expand_log_file_path(
            settings.mpm_logging_file,
        )
        log_file_level = settings.mpm_logging_file_level
        log_file_overwrite = settings.mpm_logging_file_overwrite

        self.file_handler = None
        self.file_formatter = None

        if log_file_path is not None:

            # Create logfile path, if possible:
            try:
                os.makedirs(os.path.dirname(log_file_path), exist_ok=True)
            except Exception:
                LOGGER.error(
                    LOGGING_DIRECTORY_CREATION_ERROR, log_file_path, exc_info=True
                )

            try:
                # Overwrite file if comet_mpm_file_overwrite:
                if log_file_overwrite:
                    self.file_handler = logging.FileHandler(log_file_path, "w+")
                else:
                    self.file_handler = logging.FileHandler(log_file_path)

                if log_file_level is None:
                    log_file_level = LogLevel(logging.DEBUG)

                self.file_handler.setLevel(log_file_level)
                logger_level = min(logger_level, log_file_level)

                self.file_formatter = logging.Formatter(FILE_MSG_FORMAT)
                self.file_handler.setFormatter(self.file_formatter)
                self.root.addHandler(self.file_handler)
            except Exception:
                LOGGER.error(
                    LOGGING_LOG_FILE_OPEN_ERROR,
                    log_file_path,
                    exc_info=True,
                )

        self.root.setLevel(logger_level)


COMET_LOGGING_CONFIG = None


def _setup_comet_mpm_logging() -> None:
    global COMET_LOGGING_CONFIG

    # Create a settings, read env variables if set and validate values
    settings = get_model(LogSettings)
    COMET_LOGGING_CONFIG = CometLoggingConfig(settings)
